//---------------------------------------------------------------------------------------
// Контроллер ТЭНа на Arduino Pro Mini 5V/16MHz
// Файл teh_161127.ino v.16.11.27
// OldBean
//---------------------------------------------------------------------------------------
// Скетч предназначен для управления мощностью нагрева ТЭНа методом модуляции плотности
// импульсов (PDM - Pulse Density Modulation) известным еще под назаванием метод Брезенхема.
// 
// Суть метода - пропуск части целых полупериодов сетевого напряжения к ТЭНу. Для сети 50 Гц
// мощность нагрева ТЭНа в % от номинальной мощности равно количеству пропущенных к ТЭНу 
// полупериодов за 1 секунду.
//
// Плотность пропущенных к ТЭНу импульсов (полупериодов) регулируется так, чтобы минимизировать 
// локальную ошибку дискретизации мощности (метод диффузии ошибки) и минимизировать постоянную
// составляющую, возникующую при таком способе модуляции.
//---------------------------------------------------------------------------------------
#include <Wire.h>
#define SLAVE_ADDRESS       0x04        // Адрес шины i2c для связи с внешним миром. Slave.
//---------------------------------------------------------------------------------------
#define COMMON_CATHODE // Если индикатор с общим анодом - необходимо закомментировать эту строку
//---------------------------------------------------------------------------------------
#ifdef COMMON_CATHODE // Для индикаторов с общим катодом
#define L_HIGH 1
#define L_LOW  0
#else                 // Для индикаторов с общим анодом
#define L_HIGH 0
#define L_LOW  1
#endif
//---------------------------------------------------------------------------------------
// Константы - состояние кнопок
#define NO_ACT              0
#define PLUS_CLICK          0b00000001  // Короткое нажатие кнопки "+"
#define MINUS_CLICK         0b00000010  // Короткое нажатие кнопки "-"
#define PLUS_PRESS          0b00000100  // Удержание кнопки "+" в нажатом состоянии
#define MINUS_PRESS         0b00001000  // Удержание кнопки "-" в нажатом состоянии
//---------------------------------------------------------------------------------------
static uint8_t mButPin = 0,             // Пин, к которому подключена кнопка "-"
               pButPin = 1,             // Пин, к которому подключена кнопка "+"
               outPin = 9,              // Пин выходного сигнала (для управления симистором)
               inPin = 2;               // Пин сигнала от детектора нуля (используется внешнее прерывание INT0)
//---------------------------------------------------------------------------------------
// Установки, связанные с индикатором, рассчитаны на трехразрядные семисегментные индикаторы
// с общими катодами (или анодами). Выводы анодов (или катодов для индикаторов с общими анодами) 
// подключены непосредственно к соответствующим пинам МК. Выводы общих катодов (общих анодов для
// индикаторов с общими анодами) подключены через резисторы 470 Ом.
//---------------------------------------------------------------------------------------
// Пины, к которым подключены сегменты индикатора - аноды (или катоды для индикаторов с общими анодами)
static uint8_t segPins[7] = {12, 16, 5, 7, 8, 13, 4};
// Пины, к которым подключены общие катоды (или аноды, для индикаторов с общими анодами)
static uint8_t digPins[3] = {15, 14, 11};
static uint8_t digit[11] = {0b00111111, // Маски сегментов цифр 0, 1, 2 и т.д.
                            0b00000110,
                            0b01011011,
                            0b01001111,
                            0b01100110,
                            0b01101101,
                            0b01111101,
                            0b00000111,
                            0b01111111,
                            0b01101111,
                            0b00000000}; // "Цифра" 10 это пробел
static int d[3]; // Сотни, десятки и единицы показываемые на разрядах 7-сегментного индикатора
static uint8_t mask = 0x01; // Маска для анализа цифр
static int seg = 0; // Номер сегмента, обрабатываемого (подствечиваемого, если нужно) в данный момент
static volatile int cntr = 0; // Счетчик для тактирования смены подствечиваемых сегментов
//---------------------------------------------------------------------------------------
static int dtBounce = 10; // Задержка (мс) для устранения дребезга контактов кнопок
static int dtPress = 300; // Задержка, после которой считается, что кнопка удерживается
static uint8_t bos = NO_ACT; // Предыдущее состояние кнопок
static unsigned long tpr; // Момент нажатия кнопки

static uint8_t power = 0; // Задаваемый уровень мощности в %
static uint8_t powerChangedFlag = 1; // Индикатор того, что значение мощности было изменено при помощи кнопок или i2c 
static uint8_t dPower = 5;  // Приращение устанавливаемой кнопками мощности при удержании кнопок нажатыми

static uint8_t pFlag = 0; // Наличие сигнала (импульса) на выходе модулятора
static unsigned long pStart; // Момент начала импульса на выходе модулятора
static unsigned long pWidth = 5; // Длительность управляющего импульса для PDM всегда 5 мс.
static int err = 0; // Ошибка дискретизации pdm-модулятора
static int pp = 1; // Условынй индикатор знака полупериода (1 - положительный, -1 - отрицательный)
static int ps = 0; // ps пропорциональна постоянной составляющей
//---------------------------------------------------------------------------------------
void Int0(); // Обработчик внешнего прерывания INT0
uint8_t GetKey(); // Возвращает состояние клавиатуры
void UpdateDigits(int val); // Подготовка данных (цифр) для визуализации
void ShowSegments(); // Подсвечивание очередной группы сегментов
void receiveData(int byteCount); // callback для принимаемых по i2c байтов
void sendData(); // callback для передаваемых по i2c байтов
//---------------------------------------------------------------------------------------
void setup() {
  Wire.begin(SLAVE_ADDRESS); // Инициализация библиотеки для шины i2c
  Wire.onReceive(receiveData);
  Wire.onRequest(sendData);

  pinMode(mButPin, INPUT); // Инициализация пинов, к которым подключены кнопки
  digitalWrite(mButPin, HIGH);
  pinMode(pButPin, INPUT);
  digitalWrite(pButPin, HIGH);

  pinMode(outPin, OUTPUT); // Иницализация пина выходного (модулированного) сигнала
  digitalWrite(outPin, LOW);

  pinMode(inPin, INPUT); // Инициализация пина для входного сигнала
  digitalWrite(inPin, HIGH);
  
  attachInterrupt(0, Int0, FALLING); // Прерывания по спаду импульсов входного сигнала
//  attachInterrupt(0, Int0, RISING); // Прерывания по фронту импульса входного сигнала

  for(int i = 0; i < 7; i++) { // Инициализация пинов, к которым подключены сегменты индикатора
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], L_LOW);
  }
  for(int i = 0; i < 3; i++) { // Инициализация пинов, к которым подключены общие катоды (аноды) индикатора
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], L_HIGH);
  }
  unsigned long st = millis() + 1000;
  d[0] = 8; d[1] = 8; d[2] = 8; // Для демонстрации работоспособности индикатора покажем все сегменты разом
  while(millis() < st) {
      ShowSegments();
  }
  d[0] = 10; d[1] = 10; d[2] = 10; // Очистим индикатор
  ShowSegments();

  // Инициализация таймера 2 для тактирования АЦП и посветки индикатора
  TCCR2A = (1<<WGM21);            // Режим CTC (сброс по совпадению)
  TCCR2B = (1<<CS21) | (1<<CS20); // Предделитель 32
  OCR2A = 98;                     // При этом значении совпадения происходят с частотой около 5 кГц.
  TIMSK2 = (1<<OCIE2A);           // Разрешить прерывание по совпадению
}
//---------------------------------------------------------------------------------------
ISR (TIMER2_COMPA_vect) { // Обработчик прерывания от таймера TIMER2
  sei(); // Пусть приоритет этого обработчика будет ниже, чем Int0()
  
  // Анализируем состояние кнопок
  uint8_t ks = GetKey();
  switch(ks) {
    case PLUS_CLICK:  if(power >= 100) power = 100;
                      else power++;
                      powerChangedFlag = 1;
                      break;
    case MINUS_CLICK: if(power == 0) power = 0;
                      else power--;
                      powerChangedFlag = 1;
                      break;
    case PLUS_PRESS:  if(power >= 100 - dPower) power = 100;
                      else power += dPower;
                      powerChangedFlag = 1;
                      break;
    case MINUS_PRESS: if(power <= dPower) power = 0;
                      else power -= dPower;
                      powerChangedFlag = 1;
                      break;
  }
  if(powerChangedFlag) // Скорректируем показания индикатора, если мощность была изменена
    UpdateDigits(power);
  
  if(++cntr >= 10) { // Каждые 2 мс переходим к посветке следующей группы сегментов
//    PORTB |= 0b00000010;
    cntr = 0; ShowSegments();
//    PORTB &= 0b11111101;
  }

  // И следим за уровнем управляющего сигнала на выходе
  unsigned long now = millis();
  // Если время управляющего импульса истекло - выключим его
  if(pFlag && now - pStart > pWidth) {
    digitalWrite(outPin, LOW);
    pFlag = 0;
  }
}
//---------------------------------------------------------------------------------------
void Int0() { // Функция - обработчик внешнего прерывания INT0
  int pulse = 0, lev = power + err; // Текущий уровень с учетом ошибки дискретизации, сделанной на предыдущем полупериоде.
  if(lev >= 50) pulse = 1; // Нужно бы подать импульс, но проверим на постоянную составляющую
  // Если pulse == 1, но подача импульса приведет к увеличению постоянной составляющей, то отложим 
  // подачу импульса на следующий полупериод. Он будет другого знака, что приведет к уменьшению
  // постоянной составляющей
  if(ps*pulse*pp > 0) pulse = 0;
  if(pulse) {  
    digitalWrite(outPin, HIGH); err = lev - 100;  // Считаем ошибку для следующего полупериода
  } else {
    digitalWrite(outPin, LOW); err = lev;  
  }
  pStart = millis(); pFlag = 1; // Фиксируем момент начала управляющего импульса (или его отсутствия)
  ps += pp*pulse; // Считаем текущую постоянную составляющую
  pp = -pp; // Следующий полупериод будет другого знака
}
//---------------------------------------------------------------------------------------
void loop() {
}
//---------------------------------------------------------------------------------------
uint8_t GetKey(void) {
  uint8_t ret, bcs = ((!digitalRead(mButPin) & 0x01) << 1) | (!digitalRead(pButPin) & 0x01); // Текущее состояние кнопок
  if(!bos && bcs) { // Произошло нажатие кнопки
    tpr = millis(); bos = bcs; // Фиксируем время и состояние кнопок
    return NO_ACT;
  }
  if(bos && !bcs) { // Произошло либо отжатие или дребезг кнопки
    if((millis() - tpr) < dtBounce) return NO_ACT; // Дребезг, однако...
    ret = bos; bos = bcs; return ret; // Отжатие...
  }
  if(bos && bcs) { // Возможно, удерживание кнопки
    if((millis() - tpr) < dtPress) return NO_ACT; // Но пока еще не ясно...
    // И все-таки это удержание кнопки в нажатом состоянии
    ret = bos << 2; tpr = millis(); return ret;
  }
  return NO_ACT;
}
//---------------------------------------------------------------------------------------
void UpdateDigits(int val) { // Подготовка данных (цифр) для визуализации
  d[2] = val/100;
  d[1] = (val - d[2]*100)/10;
  d[0] = val - d[2]*100 - d[1]*10;
  if(d[2] == 0) { // Не отображаем ведущие нули ежели что... ("Цифра" 10 - это "пустая" цифра, все ее сегменты потушены)
    d[2] = 10;
    if(d[1] == 0) d[1] = 10;
  }
}
//---------------------------------------------------------------------------------------
void ShowSegments() { // Подсвечивание очередной группы сегментов
  // Комментарии написаны для индикаторов с общими катодами. В скобках - для индикаторов с общими анодами
  digitalWrite(segPins[seg], L_LOW); // Опускаем (поднимаем) потенциал анода (катода) "отработавшего" сегмента
  if(++seg < 7) mask <<= 1; // Переходим к следующему сегменту
  else { // Дошли до последнего сегмента
    seg = 0; mask = 0x01;
  }
  digitalWrite(segPins[seg], L_HIGH); // Поднимаем (роняем) потенциал анода (катода) нового текущего сегмента
  for(int bi = 0; bi < 3; bi++) { // Теперь пробегаем по цифрам индикатора
    if(digit[d[bi]] & mask) // Если данный сегмент в bi-ом разряде должен гореть,
      digitalWrite(digPins[bi], L_LOW); // то опустим (поднимем) потенциал катода (анода) соответствующей цифры
    else // А если гореть не должен, то
      digitalWrite(digPins[bi], L_HIGH); // потенциал общего катода (анода) bi-й цифры должен быть высокий (низкий)
  }
}
//---------------------------------------------------------------------------------------
void receiveData(int byteCount) { // callback для принимаемых по i2c байтов
  while(Wire.available()) {
    power = Wire.read(); powerChangedFlag = 1;
  }
}
//---------------------------------------------------------------------------------------
void sendData() { // callback для передаваемых по i2c байтов
    Wire.write(power);
}
//---------------------------------------------------------------------------------------

